/*****************************************************************************
 *** Mauku - Micro-blogging client for Maemo devices
 ***
 *** Copyright (c) 2009 Henrik Hedberg <hhedberg@innologies.fi>
 ***
 *** Licensed under the Apache License, Version 2.0 (the "License");
 *** you may not use this file except in compliance with the License.
 *** You may obtain a copy of the License at
 ***
 ***     http://www.apache.org/licenses/LICENSE-2.0
 ***
 *** Unless required by applicable law or agreed to in writing, software
 *** distributed under the License is distributed on an "AS IS" BASIS,
 *** WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *** See the License for the specific language governing permissions and
 *** limitations under the License.
 ***
 *****************************************************************************/

#include "microfeedthread.h"
#include "microfeedmisc.h"
#include "microfeedstore.h"

#include <pthread.h>
#include <signal.h>

struct _MicrofeedThread {
	MicrofeedThread* next;
	MicrofeedThread* previous;

	unsigned int reference_count;
	MicrofeedMutex* mutex;
	unsigned long id;

	MicrofeedThreadPool* thread_pool;
	void* thread_implementation;
	MicrofeedThreadFunction function;
	void* in_data;
	void* out_data;
	MicrofeedThreadExitCallback exit_callback;
	void* user_data;
};

struct _MicrofeedThreadPool {
	unsigned int started_threads;
	unsigned int max_threads;
	MicrofeedThreadExitCallback exit_callback;
	void* user_data;
	MicrofeedMutex* mutex;
	MicrofeedStore* waiting_threads;
};

struct _MicrofeedMutex {
	void* mutex_implementation;
};

static void* default_thread_new(MicrofeedThreadFunction function, void* data);
static void default_thread_free(void* thread_implementation);
static void* default_thread_get_current(void);
static void default_thread_send_signal(void* thread_implementation, int signal_number);
static void default_thread_join(void* thread_implementation);
static void* default_mutex_new(void);
static void default_mutex_free(void* mutex_implementation);
static void default_mutex_lock(void* mutex_implementation);
static void default_mutex_unlock(void* mutex_implementation);
static void* thread_function(void* data);

MicrofeedThreadFunctions functions = {
	default_thread_new,
	default_thread_free,
	default_thread_get_current,
	default_thread_send_signal,
	default_thread_join,
	default_mutex_new,
	default_mutex_free,
	default_mutex_lock,
	default_mutex_unlock
};

static MicrofeedThread* threads = NULL;
static MicrofeedMutex* threads_mutex = NULL;
static unsigned long thread_count = 0;

/**
 * Sets the functions that implements the thread and mutex operations.
 * 
 * @param thread_functions A structure containing new implementions.
 */
void microfeed_thread_set_functions(MicrofeedThreadFunctions* thread_functions) {
	functions = *thread_functions;
}

void microfeed_thread_init(void) {
	if (!threads_mutex) {
		threads_mutex = microfeed_mutex_new();
	}
	if (!threads) {
		threads = microfeed_thread_get_current();
	}
}

void microfeed_thread_cleanup(void) {
	void* current_thread_implementation;
	MicrofeedThread* thread;
	
	microfeed_mutex_lock(threads_mutex);

	current_thread_implementation = functions.thread_get_current();	

	do {
		for (thread = threads; thread; thread = thread->next) {
			if (thread->thread_implementation && thread->thread_implementation != current_thread_implementation) {
				break;
			}
		}
		if (thread) {
			microfeed_mutex_unlock(threads_mutex);
			
			microfeed_thread_join(thread);
			
			microfeed_mutex_lock(threads_mutex);
		}
	} while (thread);
	
	microfeed_mutex_unlock(threads_mutex);
}

MicrofeedThread* microfeed_thread_new(MicrofeedThreadFunction function, void* data) {
	
	return microfeed_thread_new_with_exit_callback(function, data, NULL, NULL);
}

MicrofeedThread* microfeed_thread_new_with_exit_callback(MicrofeedThreadFunction function, void* data, MicrofeedThreadExitCallback exit_callback, void* user_data) {
	MicrofeedThread* thread;
	
	thread = microfeed_memory_allocate(MicrofeedThread);
	thread->reference_count = 2;
	thread->mutex = microfeed_mutex_new();
	thread->function = function;
	thread->in_data = data;
	thread->out_data= NULL;
	thread->exit_callback = exit_callback;
	thread->user_data = user_data;

	microfeed_mutex_lock(threads_mutex);
	
	thread->id = ++thread_count;

	thread->next = threads;
	if (thread->next) {
		thread->next->previous = thread;
	}
	threads = thread;
			
	microfeed_mutex_lock(thread->mutex);

	thread->thread_implementation = functions.thread_new(thread_function, thread);

	microfeed_mutex_unlock(thread->mutex);

	microfeed_mutex_unlock(threads_mutex);
	
	return thread;
}

void microfeed_thread_free(MicrofeedThread* thread) {
	functions.thread_free(thread->thread_implementation);

	microfeed_mutex_lock(threads_mutex);

	if (thread->previous) {
		thread->previous->next = thread->next;
	} else if (threads) {
		threads = thread->next;
	}
	if (thread->next) {
		thread->next->previous = thread->previous;
	}

	microfeed_mutex_unlock(threads_mutex);

	microfeed_memory_free(thread);
}

MicrofeedThread* microfeed_thread_ref(MicrofeedThread* thread) {
	microfeed_mutex_lock(thread->mutex);

	thread->reference_count++;

	microfeed_mutex_unlock(thread->mutex);
	
	return thread;
}

void microfeed_thread_unref(MicrofeedThread* thread) {
	microfeed_mutex_lock(thread->mutex);

	thread->reference_count--;
	if (thread->reference_count == 0) {
		microfeed_thread_free(thread);
	} else {
		
		microfeed_mutex_unlock(thread->mutex);
	}
}

MicrofeedThread* microfeed_thread_get_current(void) {
	MicrofeedThread* thread = NULL;
	void* thread_implementation;
	
	microfeed_mutex_lock(threads_mutex);
	
	if ((thread_implementation = functions.thread_get_current())) {
		for (thread = threads; thread; thread = thread->next) {
			if (thread->thread_implementation == thread_implementation) {
				break;
			}
		}
	}
	if (!thread) {
		thread = microfeed_memory_allocate(MicrofeedThread);
		thread->reference_count = 1;
		thread->mutex = microfeed_mutex_new();
		thread->id = 0;
		thread->function = NULL;
		thread->in_data = NULL;
		thread->out_data= NULL;

		thread->next = threads;
		if (thread->next) {
			thread->next->previous = thread;
		}
		threads = thread;
		thread->thread_implementation = thread_implementation;
	}
	
	microfeed_mutex_unlock(threads_mutex);

	return thread;
}

void microfeed_thread_send_signal(MicrofeedThread* thread, int signal_number) {
	functions.thread_send_signal(thread->thread_implementation, signal_number);
}

void microfeed_thread_join(MicrofeedThread* thread) {
	functions.thread_join(thread->thread_implementation);
}

unsigned long microfeed_thread_get_id(MicrofeedThread* thread) {
	
	return thread->id;	
}


MicrofeedThreadPool* microfeed_thread_pool_new(unsigned int maximum_thread_count) {

	return microfeed_thread_pool_new_with_exit_callback(maximum_thread_count, NULL, NULL);
}

/**
 * ...
 * 
 * An exit function is called when a <b>physical thread</b> exits.
 */
MicrofeedThreadPool* microfeed_thread_pool_new_with_exit_callback(unsigned int maximum_thread_count, MicrofeedThreadExitCallback exit_callback, void* user_data) {
	MicrofeedThreadPool* thread_pool;
	
	thread_pool = microfeed_memory_allocate(MicrofeedThreadPool);
	thread_pool->max_threads = maximum_thread_count;
	thread_pool->exit_callback = exit_callback;
	thread_pool->user_data = user_data;
	thread_pool->mutex = microfeed_mutex_new();
	thread_pool->waiting_threads = microfeed_store_new_unsorted(microfeed_store_compare_keys_direct, microfeed_store_get_key_direct);
	
	return thread_pool;
	
}

MicrofeedThread* microfeed_thread_pool_queue_thread(MicrofeedThreadPool* thread_pool, MicrofeedThreadFunction function, void* data) {

	return microfeed_thread_pool_queue_thread_with_exit_callback(thread_pool, function, data, NULL, NULL);
}

MicrofeedThread* microfeed_thread_pool_queue_thread_with_exit_callback(MicrofeedThreadPool* thread_pool, MicrofeedThreadFunction function, void* data, MicrofeedThreadExitCallback exit_callback, void* user_data) {
	MicrofeedThread* thread;
	
	thread = microfeed_memory_allocate(MicrofeedThread);
	thread->reference_count = 2;
	thread->mutex = microfeed_mutex_new();
	thread->thread_pool = thread_pool;
	thread->function = function;
	thread->in_data = data;
	thread->out_data= NULL;
	thread->exit_callback = exit_callback;
	thread->user_data = user_data;

	microfeed_mutex_lock(thread_pool->mutex);

	if (thread_pool->started_threads < thread_pool->max_threads) {
		thread_pool->started_threads++;
		
		microfeed_mutex_lock(threads_mutex);
	
		thread->next = threads;
		if (thread->next) {
			thread->next->previous = thread;
		}
		threads = thread;
				
		microfeed_mutex_lock(thread->mutex);
	
		thread->thread_implementation = functions.thread_new(thread_function, thread);
	
		microfeed_mutex_unlock(thread->mutex);
	
		microfeed_mutex_unlock(threads_mutex);
	} else {
		microfeed_store_insert(thread_pool->waiting_threads, thread);
	}		

	microfeed_mutex_unlock(thread_pool->mutex);
	
	return thread;
}

unsigned int microfeed_thread_pool_get_started_thread_count(MicrofeedThreadPool* thread_pool) {
	
	return thread_pool->started_threads;
}

unsigned int microfeed_thread_pool_get_waiting_thread_count(MicrofeedThreadPool* thread_pool) {
	
	return microfeed_store_get_size(thread_pool->waiting_threads);
}

void microfeed_thread_pool_set_maximum_thread_count(MicrofeedThreadPool* thread_pool, unsigned int maximum_thread_count) {
	
	thread_pool->max_threads = maximum_thread_count;
}

MicrofeedMutex* microfeed_mutex_new(void) {
	MicrofeedMutex* mutex;
	
	mutex = microfeed_memory_allocate(MicrofeedMutex);
	mutex->mutex_implementation = functions.mutex_new();
	
	return mutex;
}

void microfeed_mutex_free(MicrofeedMutex* mutex) {
	functions.mutex_free(mutex->mutex_implementation);
	microfeed_memory_free(mutex);
}

void microfeed_mutex_lock(MicrofeedMutex* mutex) {
	functions.mutex_lock(mutex->mutex_implementation);
}

void microfeed_mutex_unlock(MicrofeedMutex* mutex) {
	functions.mutex_unlock(mutex->mutex_implementation);
}

static void* default_thread_new(MicrofeedThreadFunction function, void* data) {
	pthread_t* thread_implementation;
	
	thread_implementation = microfeed_memory_allocate(pthread_t);
	pthread_create(thread_implementation, NULL, (void* (*)(void*))function, data);

	return (void*)thread_implementation;
}

static void default_thread_free(void* thread_implementation) {
	microfeed_memory_free(thread_implementation);
}

static void* default_thread_get_current(void) {
	pthread_t self;
	MicrofeedThread* thread;
	pthread_t* thread_implementation;
	
	self = pthread_self();
	for (thread = threads; thread; thread = thread->next) {
		if (*((pthread_t*)thread->thread_implementation) == self) {
			break;
		}
	}
	if (thread) {
		thread_implementation = thread->thread_implementation;
	} else {
		thread_implementation = microfeed_memory_allocate(pthread_t);
		*thread_implementation = self;
	}
	
	return (void*)thread_implementation;
}

static void default_thread_send_signal(void* thread_implementation, int signal_number) {
	pthread_kill(*(pthread_t*)thread_implementation, signal_number);
}

static void default_thread_join(void* thread_implementation) {
	pthread_join(*(pthread_t*)thread_implementation, NULL);
}

static void* default_mutex_new(void) {
	pthread_mutex_t* mutex_implementation;
	
	mutex_implementation = microfeed_memory_allocate(pthread_mutex_t);
	pthread_mutex_init(mutex_implementation, NULL);
	
	return (void*)mutex_implementation;
}

static void default_mutex_free(void* mutex_implementation) {
	pthread_mutex_destroy((pthread_mutex_t*)mutex_implementation);
	microfeed_memory_free(mutex_implementation);
}

static void default_mutex_lock(void* mutex_implementation) {
	pthread_mutex_lock((pthread_mutex_t*)mutex_implementation);
}

static void default_mutex_unlock(void* mutex_implementation) {
	pthread_mutex_unlock((pthread_mutex_t*)mutex_implementation);
}

static void* thread_function(void* data) {
	MicrofeedThread* thread;
	MicrofeedThreadPool* thread_pool;
	
	thread = (MicrofeedThread*)data;
	thread_pool = thread->thread_pool;

	do {
		thread->out_data = thread->function(thread->in_data);
	
		if (thread->exit_callback) {
			thread->exit_callback(thread, thread->user_data);
		}
	
		microfeed_mutex_lock(thread->mutex);
	
		thread->reference_count--;
		if (thread->reference_count == 0) {
			microfeed_thread_free(thread);
		} else {
			thread->thread_implementation = NULL;
		
			microfeed_mutex_unlock(thread->mutex);
		}
		thread = NULL;
		
		if (thread_pool) {
			microfeed_mutex_lock(thread_pool->mutex);

			if (thread_pool->started_threads <= thread_pool->max_threads && microfeed_store_get_size(thread_pool->waiting_threads) > 0) {
				thread = microfeed_store_remove_index(thread_pool->waiting_threads, 0, MicrofeedThread);	
			} else {
				thread_pool->started_threads--;
				if (thread_pool->exit_callback) {
					thread_pool->exit_callback(thread, thread_pool->user_data);
				}
			}
			
			microfeed_mutex_unlock(thread_pool->mutex);
		}
	} while (thread);

	return NULL;	
}
